/* MaskBuffer buffers geometry (for now only lines and rects),
 * renders those primitives into a mask in one go and composites
 * using the mask as geometry-buffer.
 * The advantage is that rendering rects can be done quite efficiently,
 * but composition scanline-by-scanline is very slow.
 * Please see the project-page for an in-depth explanation of MaskBuffer.
 */

#ifndef MaskBuffer_h_Included
#define MaskBuffer_h_Included

#include "jni.h"
#include <X11/Xlib.h>
#include "ArrayList.h"
#include "ArrayList.h"
#include <malloc.h>
#include <string.h>
#include <X11/extensions/Xrender.h>
#include <assert.h>

/* To not alloc/dealloc the mask pixmap tiling is used.
   This specifies the size of the alpha-tile.        */
#define MASK_TILE_SIZE 256

typedef struct {
  jint x;
  jint y;
  jint x2;
  jint y2;
}MaskRegion;

/*Holds information which belongs to a single tile*/
typedef struct {
  XrArrayList rects;
  XrArrayList lines;

  MaskRegion dirtyRectArea;
  MaskRegion dirtyLineArea;
} XrTile;

/* The MaskBuffer data-structure itself is shared in the application
   and contains pixmaps for gradient-caching and MaskFill/MaskBlit */
typedef struct {
  Pixmap lineMaskPixmap;
  Picture lineMaskPicture;
  GC drawLineGC;
  GC clearLineGC;

  Pixmap maskPixmap;
  Picture maskPicture;
  GC maskGC;
  jfloat validatedGCAlpha;

  /*Picture img1Mask;
  Picture img8Mask;
  Picture img64Mask;*/
  Picture maxMask;
  jint maxWidth;
  jint maxHeight;

  XrArrayList rects;
  XrArrayList lines;
  MaskRegion region;

  int xTiles;
  int yTiles;
  int xTileCap;
  int yTileCap;
  XrTile *tiles;

  XImage *alphaImg;
  unsigned char* alphaData;
  jint aImgWidth;
  jint aImgHeight;
}MaskBuffer;

/* Initializes the MaskBuffer data structure*/
MaskBuffer* initMaskBuffer(Window window);

/* Adds a rect which will be filled to the geometry-buffer*/ 
void addRect(MaskBuffer* buf, short x, short y, unsigned short w, unsigned short h);

/* Adds a line to the geometry-buffer*/
void addLine(MaskBuffer* buf, int x1, int y1, int x2, int y2);

/* Composites using the stored geometry as mask, using properties stored in SrcSurfaceData*/
void fillMask(MaskBuffer* buf, Picture dest);

/* Builds a 2D array of tiles and adds the rects buffered by addRect,
   to the tiles which are affacted. */
static void storeRectsInTiles(MaskBuffer *buf);

/* Builds a 2D array of tiles and adds the lines stored by addLine 
 * to the tiles which are affacted*/
static void storeLinesInTiles(MaskBuffer *buf);

/* Re-sets the MaskBuffer to an empty state */
static void cleanMaskBuffer(MaskBuffer *buf);

/* Shrinks the rectangle's coordinates if in the middle of two tiles*/
static void limitRetcCoords(XRectangle *rect);

/* Re-Sets the values of a MaskRegion */
static void initRegion(MaskRegion *region);

/* If parameters are outside the bounds of the MaskRegion, the MaskRegion will be enlarged */
static void growDirtyRegion(MaskRegion *region, jint x, jint y, jint x2, jint y2);

/* If parameters are outside the bounds of the MaskRegion, the MaskRegion will be enlarged, 
   but is limited to the size of a single tile */
static void growDirtyRegionTileLimit(MaskRegion *region, jint x, jint y, jint x2, jint y2);

/* Re-Sets a tile data-structure*/
static void initXrTile(XrTile *tile);

/* Returns the total dirty size of both, the rect and the line tile combined */
static void getTileDirtyArea(MaskRegion *dirtyRectArea, MaskRegion *dirtyLineArea, MaskRegion *totalDirtyArea);

/* Translates all rects in rectList by x/y */
static void translateRects(XrArrayList *rectList, int x, int y);

/* Allocates and initializes the tile-datastructures nescessary
   to hold the geometry stored in the MaskBuffer*/
static void allocTileBuffer(MaskBuffer* buf);

#endif /* MaskBuffer_h_Included */
