/*
 * Copyright 2000-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include "X11SurfaceData.h"
#include "XRSurfaceData.h"
#include "SpanIterator.h"
#include "Trace.h"
#include "ProcessPath.h"
#include "GraphicsPrimitiveMgr.h"
#include <math.h>
#include <stdint.h>

#include <jlong.h>

#define POLYTEMPSIZE    (int)(256 / sizeof(XPoint))
#define ABS(n)          (((n) < 0) ? -(n) : (n))

#define MAX_SHORT 32767
#define MIN_SHORT (-32768)

#define CLAMP_TO_SHORT(x) (((x) > MAX_SHORT)                            \
                           ? MAX_SHORT                                  \
                           : ((x) < MIN_SHORT)                          \
                               ? MIN_SHORT                              \
                               : (x))

#define CLAMP_TO_USHORT(x)  (((x) > 65535) ? 65535 : ((x) < 0) ? 0 : (x))


/* Adds a PolyLine consisting of multiple points to the MaskBuffer */
JNIEXPORT void JNICALL XRDrawLines(Picture dst, XPoint * points, int npoints)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_XRDrawLines");

    int i;
    XPoint *start = &points[0], *end;
    for(i=1; i < npoints; i++)
    {
      end = &points[i];
      addLine(maskBuffer, start->x, start->y, end->x, end->y);
      start = end;
    }
#endif
}

/*
 * Class:     sun_java2d_xr_XRRenderer
 * Method:    XDrawLine
 * Signature: (IJIIII)V
 */
JNIEXPORT void JNICALL Java_sun_java2d_xr_XRRenderer_XRDrawLine
    (JNIEnv *env, jobject xr, jlong pXSData, jint x1, jint y1, jint x2, jint y2)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_XRDrawLine");
    X11SDOps *xsdo = (X11SDOps *) jlong_to_ptr(pXSData);

    if (xsdo == NULL) {
        return;
    }

    addLine(maskBuffer, x1, y1, x2, y2);
    fillMask(maskBuffer, xsdo->xrPic);

    X11SD_DirectRenderNotify(env, xsdo);
#endif /* !HEADLESS */
}

static void storeLine(DrawHandler* hnd,
                      jint x0, jint y0, jint x1, jint y1)
{
#ifndef HEADLESS
   addLine(maskBuffer, x0, y0, x1, y1);
#endif /* !HEADLESS */
}

static void storePoint(DrawHandler* hnd, jint x0, jint y0) {
#ifndef HEADLESS
    addRect(maskBuffer, x0, y0, 1, 1);
#endif /* !HEADLESS */
}

static void drawSubPath(ProcessHandler* hnd) {
#ifndef HEADLESS
#endif /* !HEADLESS */
}

static void drawScanline(DrawHandler* hnd, jint x0, jint x1, jint y0)
{
#ifndef HEADLESS
   addRect(maskBuffer, x0, y0, x1-x0 + 1, 1);
#endif /* !HEADLESS */
}

/*
 * Class:     sun_java2d_xr_XRRenderer
 * Method:    XDoPath
 * Signature: (Lsun/java2d/SunGraphics2D;JJIILjava/awt/geom/Path2D/Float;Z)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRRenderer_XRDoPath
    (JNIEnv *env, jobject self, jobject sg2d, jlong pXSData,
     jint transX, jint transY, jobject p2df, jboolean isFill)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_XRDoPath");

    X11SDOps *xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    jarray typesArray;
    jobject pointArray;
    jarray coordsArray;
    jint numTypes;
    jint fillRule;
    jint maxCoords;
    jbyte *types;
    jfloat *coords;

    DrawHandler drawHandler = {
        NULL, NULL, NULL,
        MIN_SHORT, MIN_SHORT, MAX_SHORT, MAX_SHORT,
        0, 0, 0, 0,
        NULL
    };
    PHStroke stroke;

    if (xsdo == NULL) {
        return;
    }

    if (isFill) {
        fillRule = (*env)->GetIntField(env, p2df, path2DWindingRuleID);
    }

    typesArray = (jarray)(*env)->GetObjectField(env, p2df, path2DTypesID);
    coordsArray = (jarray)(*env)->GetObjectField(env, p2df,
                                                 path2DFloatCoordsID);
    if (coordsArray == NULL) {
        JNU_ThrowNullPointerException(env, "coordinates array");
        return;
    }
    numTypes = (*env)->GetIntField(env, p2df, path2DNumTypesID);
    if ((*env)->GetArrayLength(env, typesArray) < numTypes) {
        JNU_ThrowArrayIndexOutOfBoundsException(env, "types array");
        return;
    }


    drawHandler.pData = NULL;

    stroke = (((*env)->GetIntField(env, sg2d, sg2dStrokeHintID) ==
               sunHints_INTVAL_STROKE_PURE)
              ? PH_STROKE_PURE
              : PH_STROKE_DEFAULT);

    maxCoords = (*env)->GetArrayLength(env, coordsArray);
    coords = (jfloat*)
        (*env)->GetPrimitiveArrayCritical(env, coordsArray, NULL);
    if (coords != NULL) {
        types = (jbyte*)
            (*env)->GetPrimitiveArrayCritical(env, typesArray, NULL);
        if (types != NULL) {
            jboolean ok;

            if (isFill) {
                drawHandler.pDrawScanline = &drawScanline;
                ok = doFillPath(&drawHandler,
                                transX, transY,
                                coords, maxCoords,
                                types, numTypes,
                                stroke, fillRule);
            } else {
                drawHandler.pDrawLine = &storeLine;
                drawHandler.pDrawPixel = &storePoint;
                ok = doDrawPath(&drawHandler, &drawSubPath,
                                transX, transY,
                                coords, maxCoords,
                                types, numTypes,
                                stroke);
            }
            if (!ok) {
                JNU_ThrowArrayIndexOutOfBoundsException(env, "coords array");
            }
            (*env)->ReleasePrimitiveArrayCritical(env, typesArray, types,
                                                  JNI_ABORT);
        }
        (*env)->ReleasePrimitiveArrayCritical(env, coordsArray, coords,
                                              JNI_ABORT);
    }


    fillMask(maskBuffer, xsdo->xrPic);

    X11SD_DirectRenderNotify(env, xsdo);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL Java_sun_java2d_xr_XRRenderer_XRFillRect
    (JNIEnv *env, jobject xr,
     jlong pXSData,
     jint x, jint y, jint w, jint h)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_XRFillRect");
    X11SDOps *xsdo = (X11SDOps *) jlong_to_ptr(pXSData);

    if (xsdo == NULL) {
        return;
    }
 
    addRect(maskBuffer, x, y, w, h);
    fillMask(maskBuffer, xsdo->xrPic);

    X11SD_DirectRenderNotify(env, xsdo);
#endif /* !HEADLESS */
}

JNIEXPORT void JNICALL Java_sun_java2d_xr_XRRenderer_XRFillSpans
    (JNIEnv *env, jobject xr,
     jlong pXSData,
     jobject si, jlong pIterator,
     jint transx, jint transy)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_XRFillSpans");

    SpanIteratorFuncs *pFuncs = (SpanIteratorFuncs *) jlong_to_ptr(pIterator);
    void *srData;
    jint x, y, w, h;
    jint spanbox[4];
    X11SDOps *xsdo = (X11SDOps *) jlong_to_ptr(pXSData);

    if (xsdo == NULL) {
        return;
    }

    if (JNU_IsNull(env, si)) {
        JNU_ThrowNullPointerException(env, "span iterator");
        return;
    }
    if (pFuncs == NULL) {
        JNU_ThrowNullPointerException(env, "native iterator not supplied");
        return;
    }

    srData = (*pFuncs->open)(env, si);
    while ((*pFuncs->nextSpan)(srData, spanbox)) {
        x = spanbox[0] + transx;
        y = spanbox[1] + transy;
        w = spanbox[2] - spanbox[0];
        h = spanbox[3] - spanbox[1];

        addRect(maskBuffer, x, y, w, h);
    }

    fillMask(maskBuffer, xsdo->xrPic);

    (*pFuncs->close)(env, srData);
    X11SD_DirectRenderNotify(env, xsdo);
#endif /* !HEADLESS */
}

/*
 * Class:     sun_java2d_xr_XRRenderer
 * Method:    devCopyArea
 * Signature: (Lsun/java2d/SurfaceData;IIIIII)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRRenderer_devCopyArea
    (JNIEnv *env, jobject xr,
     jlong xsd, jlong gc,
     jint srcx, jint srcy,
     jint dstx, jint dsty,
     jint width, jint height)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRRenderer_devCopyArea");

    X11SDOps *xsdo;
    GC xgc;

    xsdo = (X11SDOps *) jlong_to_ptr(xsd);
    if (xsdo == NULL) {
        return;
    }

    xgc = (GC) jlong_to_ptr(gc);
    if (xgc == NULL) {
        return;
    }

    XCopyArea(awt_display, xsdo->drawable, xsdo->drawable, xgc,
              srcx, srcy, width, height, dstx, dsty);

    X11SD_DirectRenderNotify(env, xsdo);
#endif /* !HEADLESS */
}
