/*
 * Copyright 2001-2003 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include "GraphicsPrimitiveMgr.h"
#include "Region.h"
#include "Trace.h"
#include "X11SurfaceData.h"

#ifndef HEADLESS
SrcSurfaceData xrSrcData;
MaskBuffer *maskBuffer;
XRenderColor color_black;
#endif /* !HEADLESS */

JNIEXPORT void JNICALL
   Java_sun_java2d_xr_XRSurfaceData_XRInitXRender(JNIEnv *env, jobject xsd, jlong pXSData, jint pictFormat)
{
#ifndef HEADLESS
  J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_initXRender");
  X11SDOps *xsdo;
  xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
  if (xsdo == NULL) {
      return;
  }

  if(xsdo->xrPic == None)
  {
    XRenderPictureAttributes pict_attr;
    pict_attr.repeat = RepeatNone; 
    XRenderPictFormat *fmt = XRenderFindStandardFormat(awt_display, pictFormat);
    xsdo->xrPic = XRenderCreatePicture(awt_display, xsdo->drawable, fmt, CPRepeat, &pict_attr);
  }

  /*TODO: Improper place to initialize that stuff here - however we need a drawable in order to create the pixmaps*/
  if(maskBuffer == NULL) {
    maskBuffer = initMaskBuffer(xsdo->drawable);
    initXRSrcData(xsdo->drawable);
  }
#endif /* !HEADLESS */
}

static void initXRSrcData(Drawable *drawable) {
#ifndef HEADLESS
    XRenderPictureAttributes pict_attr;
    pict_attr.repeat = 1;

    /*Init solid color pen*/
    XRenderPictFormat *fmt32 = XRenderFindStandardFormat(awt_display, PictStandardARGB32);
    Pixmap solidPixmap = XCreatePixmap(awt_display, drawable, 1, 1, 32);
    xrSrcData.solid = XRenderCreatePicture(awt_display, solidPixmap, fmt32, CPRepeat, &pict_attr);

    /*Init extra-alpha mask*/
    XRenderPictFormat *fmt8 = XRenderFindStandardFormat(awt_display, PictStandardA8);
    Pixmap alphaPixmap = XCreatePixmap(awt_display, drawable, 1, 1, 8);
    xrSrcData.alphaMaskPict = XRenderCreatePicture(awt_display, alphaPixmap, fmt8, CPRepeat, &pict_attr);

    /*Init gradient cache*/
    if(xrSrcData.gradCacheEnabled) {
      xrSrcData.gradPixmap = XCreatePixmap(awt_display, drawable, MASK_TILE_SIZE, MASK_TILE_SIZE, 32);
      xrSrcData.gradPict = XRenderCreatePicture(awt_display, xrSrcData.gradPixmap, fmt32, 0, &pict_attr);
    }
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_initIDs(JNIEnv *env, jclass xsd, jboolean gradCacheEnabled)
{
#ifndef HEADLESS
  J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_initIDs");

  xrSrcData.src = None;
  xrSrcData.solid = None;
  xrSrcData.texture = None;
  xrSrcData.gradient = None;
  xrSrcData.alphaMask = None;
  xrSrcData.alphaMaskPict = None;
  xrSrcData.compRule = PictOpOver;
  xrSrcData.extraAlpha = 1.0f;
  xrSrcData.gradCacheEnabled = gradCacheEnabled;
  maskBuffer = NULL;

  color_black.red=0;
  color_black.green=0;
  color_black.blue=0;
  color_black.alpha=0xffff;

  XShared_initIDs(env, JNI_FALSE);
#endif /* !HEADLESS */
}


void XRComposite(Picture mask, Picture dest, int srcX, int srcY, int maskX, int maskY, int dstX, int dstY, unsigned int width, unsigned int height)
{ 
  Picture cachedSrc = xrSrcData.src;
  int cachedX = srcX;
  int cachedY = srcY;

  if(xrSrcData.src == xrSrcData.gradient && xrSrcData.gradCacheEnabled)
  {
    XRenderComposite(awt_display, PictOpSrc, xrSrcData.gradient, None, xrSrcData.gradPict, srcX, srcY, 0, 0, 0, 0, width, height);
    cachedX = 0;
    cachedY = 0; 
    cachedSrc = xrSrcData.gradPict;
  }

   XRenderComposite (awt_display, xrSrcData.compRule, cachedSrc, mask, dest, cachedX, cachedY, maskX, maskY, dstX, dstY, width, height);
}



JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRInitSurface(JNIEnv *env, jclass xsd,
                                               jint depth,
                                               jint width, jint height,
                                               jlong drawable, jint pictFormat)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_initSurface");

    jboolean success = JNI_FALSE;
    X11SDOps *xsdo = X11SurfaceData_GetOps(env, xsd);
    if (xsdo == NULL) {
        return;
    }

    if(XShared_initSurface(env, xsdo, depth, width, height, drawable))
    {
      XRenderPictureAttributes pict_attr;
      XRenderPictFormat *fmt = XRenderFindStandardFormat(awt_display, pictFormat); 
      xsdo->xrPic = XRenderCreatePicture(awt_display, xsdo->drawable, fmt, 0, &pict_attr);
    }
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRResetPaint
    (JNIEnv *env, jclass xsd)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRResetPaint");
    xrSrcData.src = xrSrcData.solid;
#endif /* !HEADLESS */
}

JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetRepeat
    (JNIEnv *env, jclass xsd, jlong pXSData, jint repeat)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetRepeat");
    X11SDOps *xsdo;
    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    XRenderPictureAttributes pict_attr;
    pict_attr.repeat = repeat;
    XRenderChangePicture (awt_display, xsdo->xrPic, CPRepeat, &pict_attr);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetLinearGradientPaint
    (JNIEnv *env, jclass xsd, jfloatArray fractionsArray, jintArray pixelsArray, jint x1, jint y1, jint x2, jint y2, jint numStops, jint cycleMethod, jboolean useMask, jboolean linear, jint m00, jint m01, jint m02, jint m10, jint m11, jint m12)
{
#ifndef HEADLESS
   J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetLinearGradientPaint");

   jint i;
   jint* pixels;
   jfloat* fractions;

   if((pixels = (jint *) (*env)->GetPrimitiveArrayCritical(env, pixelsArray, NULL)) == NULL) {
     return;
   }
   if((fractions = (jfloat *) (*env)->GetPrimitiveArrayCritical(env, fractionsArray, NULL)) == NULL) {
     return; //TODO release pixels first
   }

   if(xrSrcData.gradient != None) {
      XRenderFreePicture(awt_display, xrSrcData.gradient);
      xrSrcData.gradient = None;
   }

    XLinearGradient grad;
    grad.p1.x = x1;
    grad.p1.y = y1;
    grad.p2.x = x2;
    grad.p2.y = y2;

    /*TODO optimized & malloc check*/
    XRenderColor *colors = (XRenderColor *) malloc(numStops * sizeof(XRenderColor));
    XFixed *stops =  (XFixed *) malloc(numStops * sizeof(XFixed));

    for(i=0; i < numStops; i++) {
      stops[i] = XDoubleToFixed(fractions[i]);
      decodeRenderColorPre(pixels[i], colors + i);
    }

    xrSrcData.gradient = XRenderCreateLinearGradient(awt_display, &grad, stops, colors, numStops);
    free(colors);
    free(stops);

    xrSrcData.src = xrSrcData.gradient;

    XTransform tr;
    BUILD_TRANSFORM_MATRIX(tr, m00, m01, m02, m10, m11, m12);
    XRenderSetPictureTransform (awt_display, xrSrcData.gradient, &tr);

    XRenderPictureAttributes pict_attr;
    pict_attr.repeat = cycleMethod;
    XRenderChangePicture (awt_display, xrSrcData.gradient, CPRepeat, &pict_attr);

   (*env)->ReleasePrimitiveArrayCritical(env, pixelsArray, pixels, JNI_ABORT);
   (*env)->ReleasePrimitiveArrayCritical(env, fractionsArray, fractions, JNI_ABORT);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetRadialGradientPaint
    (JNIEnv *env, jclass xsd, jfloatArray fractionsArray, jintArray pixelsArray, jint fx, jint numStops, jint cycleMethod, jboolean useMask, jboolean linear, jint innerRadius, jint outerRadius, jint m00, jint m01, jint m02, jint m10, jint m11, jint m12)
{
#ifndef HEADLESS
   J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetRadialGradientPaint");
   jint i;
   jint* pixels;
   jfloat* fractions;

   if((pixels = (jint *) (*env)->GetPrimitiveArrayCritical(env, pixelsArray, NULL)) == NULL) {
     return;
   }
   if((fractions = (jfloat *) (*env)->GetPrimitiveArrayCritical(env, fractionsArray, NULL)) == NULL) {
     return;
   }

   if(xrSrcData.gradient != None) {
      XRenderFreePicture(awt_display, xrSrcData.gradient);
      xrSrcData.gradient = None;
   }

    XRadialGradient grad;
    grad.inner.x = 0;
    grad.inner.y = 0;
    grad.inner.radius = innerRadius;
    grad.outer.x = 0;
    grad.outer.y = 0;
    grad.outer.radius = outerRadius;

    /*TODO optimized & malloc check*/
    XRenderColor *colors = (XRenderColor *) malloc(numStops * sizeof(XRenderColor));
    XFixed *stops = (XFixed *) malloc(numStops * sizeof(XFixed));

    for(i=0; i < numStops; i++) {
      stops[i] = XDoubleToFixed(fractions[i]);
      decodeRenderColorPre(pixels[i], colors + i);
    }

    xrSrcData.gradient = XRenderCreateRadialGradient(awt_display, &grad, stops, colors, numStops);
    free(colors);
    free(stops);

    xrSrcData.src = xrSrcData.gradient;

    XTransform tr;
    BUILD_TRANSFORM_MATRIX(tr, m00, m01, m02, m10, m11, m12);
    XRenderSetPictureTransform (awt_display, xrSrcData.gradient, &tr);

    XRenderPictureAttributes pict_attr;
    pict_attr.repeat = cycleMethod;
    XRenderChangePicture (awt_display, xrSrcData.gradient, CPRepeat, &pict_attr);

   (*env)->ReleasePrimitiveArrayCritical(env, pixelsArray, pixels, JNI_ABORT);
   (*env)->ReleasePrimitiveArrayCritical(env, fractionsArray, fractions, JNI_ABORT);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetTexturePaint
    (JNIEnv *env, jclass xsd, jlong pXSData)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetTexturePaint");

    X11SDOps *xsdo;
    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    xrSrcData.texture = xsdo->xrPic;
    xrSrcData.src = xrSrcData.texture;
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetComposite
    (JNIEnv *env, jclass xsd, jint compRule, jfloat eAlpha)
{
#ifndef HEADLESS
  J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetComposite");

  xrSrcData.compRule = compRule;
  xrSrcData.extraAlpha = eAlpha;

  if(eAlpha != 1.0f) {
    unsigned short alphaValue = XDoubleToUShort(eAlpha);
    XRenderColor alphaColor;
    alphaColor.alpha = alphaValue;
    alphaColor.red = alphaValue;
    alphaColor.green = alphaValue;
    alphaColor.blue = alphaValue;
    XRenderFillRectangle (awt_display, PictOpSrc, xrSrcData.alphaMaskPict, &alphaColor, 0, 0, 1, 1);
    xrSrcData.alphaMask = xrSrcData.alphaMaskPict;
  }else {
    xrSrcData.alphaMask = None;
  }
#endif /* !HEADLESS */
}

JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetClip
    (JNIEnv *env, jclass xsd, jlong pXSData,
     jint x1, jint y1, jint x2, jint y2,
     jobject complexclip)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetClip");

    int numrects;
    XRectangle rects[256];
    XRectangle *pRect = rects;
    X11SDOps *xsdo;

    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    numrects = RegionToYXBandedRectangles(env,
            x1, y1, x2, y2, complexclip,
            &pRect, 256);


     XRenderSetPictureClipRectangles (awt_display, xsdo->xrPic, 0, 0, pRect, numrects);

    if (pRect != rects) {
        free(pRect);
    }
#endif /* !HEADLESS */
}

JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRResetClip
    (JNIEnv *env, jclass xsd, jlong pXSData)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRResetClip");
    X11SDOps *xsdo;

    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    XRectangle clip;
    clip.x = 0;
    clip.y = 0;
    clip.width = 32767;
    clip.height = 32767;

    XRenderSetPictureClipRectangles (awt_display, xsdo->xrPic, 0, 0, &clip, 1);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetTransformNative
    (JNIEnv *env, jclass xsd, jlong pXSData, jint m00, jint m01, jint m02, jint m10, jint m11, jint m12)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetTransformNative");
    X11SDOps *xsdo;

    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    XTransform tr;
    BUILD_TRANSFORM_MATRIX(tr, m00, m01, m02, m10, m11, m12);

    XRenderSetPictureTransform (awt_display, xsdo->xrPic, &tr);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetForeground
    (JNIEnv *env, jclass xsd, jint pixel)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetForeground");

    decodeRenderColor(pixel, &xrSrcData.solidColor);
    XRenderFillRectangle (awt_display, PictOpSrc, xrSrcData.solid, &xrSrcData.solidColor, 0, 0, 1, 1);
#endif /* !HEADLESS */
}


JNIEXPORT void JNICALL
Java_sun_java2d_xr_XRSurfaceData_XRSetFilter
    (JNIEnv *env, jclass xsd, jlong pXSData, jint filter)
{
#ifndef HEADLESS
    J2dTraceLn(J2D_TRACE_INFO, "in XRSurfaceData_XRSetFilter");
    X11SDOps *xsdo;

    xsdo = (X11SDOps *) jlong_to_ptr(pXSData);
    if (xsdo == NULL) {
        return;
    }

    switch(filter)
    {
      case 0:
        XRenderSetPictureFilter(awt_display, xsdo->xrPic, "fast", NULL, 0);
        break;

      case 1:
        XRenderSetPictureFilter(awt_display, xsdo->xrPic, "good", NULL, 0);
        break;

      case 2:
        XRenderSetPictureFilter(awt_display, xsdo->xrPic, "best", NULL, 0);
        break;
    }
#endif /* !HEADLESS */
}

static void decodeRenderColor(jint pixel, XRenderColor *color)
{ 
   /*Extract color values from pixel, premultiply values. TODO: shouldn't this be done at the Java-level?*/
    XFixed alpha = (((pixel & 0xFF000000) >> 16) + 255);
    XFixed red =   (((pixel & 0x00FF0000) >> 8)  + 255);
    XFixed green = (((pixel & 0x0000FF00) >> 0)  + 255);
    XFixed blue =  (((pixel & 0x000000FF) << 8)  + 255);

    /* Optimizations are based on the fact that alpha is 0xffff,
     * but XRender maybe also optimizes if alpha==0, so we correct here in that case */
    if(alpha == 255) {
      alpha = 0;
    }

    XDouble alphaMult = XFixedToDouble(alpha);
    color->alpha = alpha;
    color->red = (XFixed) red*alphaMult;
    color->green = (XFixed) green*alphaMult;
    color->blue = (XFixed) blue*alphaMult;
}

static void decodeRenderColorPre(jint pixel, XRenderColor *color)
{ 
    color->alpha = (((pixel & 0xFF000000) >> 16) + 255); /*REMIND: We rely on optimizations that this is 0xffff for opaque colors*/
    color->red =   (((pixel & 0x00FF0000) >> 8)  + 255);
    color->green = (((pixel & 0x0000FF00) >> 0)  + 255);
    color->blue =  (((pixel & 0x000000FF) << 8)  + 255);

    if(color->alpha == 255) {
      color->alpha = 0;
    }
}